#include <algorithm>
#include <cmath>
#include <queue>
#include <vector>

#include <SDL_pixels.h>

#include <basic_widgets/core/draw.hpp>
#include <basic_widgets/core/font.hpp>
#include <basic_widgets/core/math.hpp>
#include <basic_widgets/core/renderer.hpp>
#include <basic_widgets/core/texture.hpp>

#include "basic_widgets/core/type/sdl_error.hpp"
#include "SDL_surface.h"

namespace bwidgets
{
    auto aa(const Color base_color, const int aa_pixels, const float d) noexcept -> Color
    {
        Color c = [aa_pixels, base_color, d]() -> Color {
            const auto [r, g, b, a] = base_color();
            // AA disabled, returns fully opaque or fully transparent colors
            if (aa_pixels == 0) {
                if (d > 0) return {r, g, b, 0};
                return {r, g, b, a};
            }
            const auto factor = 1.0 - smoothstep<float>(d, (float)-aa_pixels, 0.);
            return {r, g, b, (uint8_t)((float)base_color().a * factor)};
        }();

        return c;
    }

    void fill_rect_gradient(Renderer& rend, const SDL_Rect& rect, const Color start,
                            const Color end, const int steps)
    {
        for (int i = 0; i < steps; i++) {
            const auto factor = linear(i, 0, steps - 1);
            rend.draw_color(lerp(start, end, factor));
            rend.fill_rect(rect_margin(rect, {i, i}));
        }
    }

    auto filled_circle(const Color c, const int resolution, const Renderer& r,
                       const int aa_pixels) -> std::shared_ptr<Texture>
    {
        if (resolution < 1) return nullptr;

        // clang-format off
        auto           texture {std::make_shared<Texture>(
                                 r,
                                 SDL_PIXELFORMAT_RGBA32,
                                 SDL_TEXTUREACCESS_STATIC,
                                 resolution,
                                 resolution)};
        // clang-format on
        const auto      radius {resolution / 2};
        const SDL_Point center {radius, radius};

        set_pixels_color(texture.get(),
                         [aa_pixels, c, center, radius](
                           SDL_Point p, const SDL_PixelFormat& format) -> uint32_t {
                             const auto d_delta  = distance(center, p) - (float)radius;
                             const auto aa_color = aa(c, aa_pixels, d_delta);

                             return SDL_MapRGBA(&format, aa_color().r, aa_color().g,
                                                aa_color().b, aa_color().a);
                         });
        texture->blend_mode(SDL_BLENDMODE_BLEND);
        texture->scale_mode(SDL_ScaleModeNearest);

        return texture;
    }

    auto render_text(const std::vector<std::string>& lines, const Color& c, Font& f,
                     const Renderer& r) -> std::shared_ptr<Texture>
    {
        int                                               width {0};
        std::queue<std::unique_ptr<SDL_Surface, Deleter>> surfaces {};
        for (const auto& l : lines) {
            auto s = f.render(Font::RenderMode::BLENDED, l, c);
            if (s->w > width) width = s->w;
            surfaces.emplace(std::move(s));
        }

        std::unique_ptr<SDL_Surface, Deleter> text_surface {
          SDL_CreateRGBSurfaceWithFormat(0, width, surfaces.size() * f.height, 32,
                                         SDL_PIXELFORMAT_RGBA32)};
        int y {0};
        int i {0};
        while (!surfaces.empty()) {
            auto&    s {surfaces.front()};
            SDL_Rect dst {0, y, s->w, s->h};
            SDL_BlitSurface(s.get(), nullptr, text_surface.get(), &dst);
            surfaces.pop();
            y += f.height;
        }

        return std::make_shared<Texture>(r, text_surface.get());
    }

    void set_pixels_color(
      Texture*                                                          t,
      const std::function<uint32_t(SDL_Point, const SDL_PixelFormat&)>& pixel_color)
    {
        auto attr  = t->attributes();
        auto pitch = attr.w * attr.format->BytesPerPixel;

        std::vector<uint32_t> pixels;
        pixels.reserve(attr.h * (std::vector<uint32_t>::size_type)pitch);
        // TODO parallel algo
        for (auto y = 0; y < attr.h; y++) {
            for (auto x = 0; x < attr.w; x++) {
                pixels.emplace_back(pixel_color({x, y}, *attr.format));
            }
        }

        t->update(nullptr, (const void*)pixels.data(), pitch);
    }
}
