#include <cmath>

#include <basic_widgets/core/draw.hpp>
#include <basic_widgets/core/math.hpp>
#include <basic_widgets/w/button_impl.hpp>
#include <basic_widgets/w/widget_factory.hpp>

using namespace bwidgets;

ButtonImpl::ButtonImpl(Widget* parent) noexcept
  : WidgetImpl {parent}, _caption {create_caption(this)}
{
    enable_mouse_handler(_mouse_area, viewport());

    _caption->alignment(Caption::Alignment::CENTER);
    _caption->render_mode(Font::RenderMode::BLENDED);
}

auto ButtonImpl::size() const noexcept -> Size
{
    return _caption->size() + theme()->size_widget_border() * 2;
}

auto ButtonImpl::text() const noexcept -> std::string_view
{
    return _caption->text();
}

void ButtonImpl::text(std::string txt)
{
    _caption->text(std::move(txt));
    _handle_geometry_change(viewport());
}

void ButtonImpl::_handle_font_change(const std::shared_ptr<Font>& f)
{
    _caption->font(f);
    _handle_geometry_change(viewport());
}

void ButtonImpl::_handle_font_color_change(const Color& fg, const Color&)
{
    _caption->font_color_fg(fg);
}

void ButtonImpl::_handle_geometry_change(const SDL_Rect& vp)
{
    _mouse_area = {0, 0, vp.w, vp.h};

    const auto txt_size = _caption->size();
    _caption_area       = {center_line(vp.w, txt_size.w), center_line(vp.h, txt_size.h),
                     txt_size.w, txt_size.h};

    _caption->viewport(rect_offset(_caption_area, vp));
}

void ButtonImpl::_handle_renderer_change(const std::shared_ptr<Renderer>& r)
{
    _caption->renderer(r);
}

void ButtonImpl::_handle_theme_change(const std::shared_ptr<Theme>& t)
{
    _caption->theme(t);
    _caption->font(t->font_default());
}

void ButtonImpl::_handle_rendering()
{
    const Color& c =
      hovered() ? theme()->color_widget_bg_hl() : theme()->color_widget_bg();

    if (pushed())
        fill_rect_gradient(*renderer(), _mouse_area,
                           theme()->color_widget_border_pushed(), c,
                           theme()->size_widget_border());
    else
        fill_rect_gradient(*renderer(), _mouse_area, theme()->color_widget_border(), c,
                           theme()->size_widget_border());

    _caption->font_color_bg(c);
    _caption->render();
}

void ButtonImpl::_on_push(const bool state)
{
    // Move slightly the caption position to give a push effect.
    const auto offset = [state, this]() -> SDL_Point {
        if (state) {
            return {viewport().x + 1, viewport().y + 1};
        }
        return {viewport().x, viewport().y};
    }();

    _caption->viewport(rect_offset(_caption_area, offset));
}
