#include <basic_widgets/core/math.hpp>
#include <basic_widgets/core/texture.hpp>
#include <basic_widgets/w/caption_impl.hpp>
#include <basic_widgets/w/numeric_input.hpp>

using namespace bwidgets;

auto CaptionImpl::alignment() const -> Alignment
{
    return _alignment;
}

void CaptionImpl::alignment(const Alignment a)
{
    _alignment = a;
}

auto CaptionImpl::render_mode() const -> Font::RenderMode
{
    return _render_mode;
}

void CaptionImpl::render_mode(const Font::RenderMode m)
{
    if (m != _render_mode) {
        _text_texture.reset();
        _render_mode = m;
    }
}

auto CaptionImpl::size() const noexcept -> Size
{
    if (!font()) return {0, 0};

    return font()->text_size(_text);
}

auto CaptionImpl::text() const noexcept -> std::string_view
{
    return _text;
}

void CaptionImpl::text(std::string t)
{
    if (t != _text) {
        _text_texture.reset();
        _text = std::move(t);
    }
}

void CaptionImpl::_handle_font_change(const std::shared_ptr<Font>&)
{
    _text_texture.reset();
}

void CaptionImpl::_handle_font_color_change(const Color& fg, const Color& bg)
{
    if (fg != font_color_fg()
        || (bg != font_color_bg() && render_mode() == Font::RenderMode::SHADED))
    {
        _text_texture.reset();
    }
}

void CaptionImpl::_handle_geometry_change(const SDL_Rect& vp)
{
    if (vp.w != viewport().w) {
        _text_area = {0, 0, vp.w, vp.h};
    }
    else if (vp.h != viewport().h) {
        _text_area = {0, 0, vp.w, vp.h};
        _text_texture.reset();
    }
}

void CaptionImpl::_handle_renderer_change(const std::shared_ptr<Renderer>&)
{
    _text_texture.reset();
}

void CaptionImpl::_handle_rendering()
{
    if (!_text_texture) _handle_texture_update();
    if (!_text_texture) throw std::runtime_error("can't render text");

    // fill caption viewport with background color when using
    // shaded text rendering mode.
    if (_render_mode == Font::RenderMode::SHADED) {
        renderer()
          ->draw_color(font_color_bg())
          ->fill_rect({0, 0, viewport().w, viewport().h});
    }

    const Size size_dst {
      (int)((float)_text_texture->attributes().w / (float)_text_texture->attributes().h
            * (float)_text_area.h),
      _text_area.h};

    const auto texture_dst = [size_dst, this]() -> SDL_Rect {
        switch (alignment()) {
            case Alignment::CENTER:
                return {center_line(_text_area.w, size_dst.w) + _text_area.x,
                        _text_area.y, size_dst.w, size_dst.h};
            case Alignment::LEFT:
                return {_text_area.x, _text_area.y, size_dst.w, size_dst.h};
            case Alignment::RIGHT:
                return {_text_area.w - size_dst.w + _text_area.x, _text_area.y,
                        size_dst.w, size_dst.h};
            default:
                throw std::logic_error("missing switch case.");
        }
    }();
    renderer()->copy(*_text_texture, nullptr, texture_dst);
}

void CaptionImpl::_handle_texture_update()
{
    if (!font()) return;
    _text_texture.reset();

    auto s = [this]() {
        switch (render_mode()) {
            case Font::RenderMode::SHADED:
                return font()->render(_render_mode, _text, font_color_fg(),
                                      font_color_bg());
            default:
                return font()->render(_render_mode, _text, font_color_fg());
        }
    }();
    _text_texture = std::make_shared<Texture>(*renderer(), s.get());
    _text_texture->alpha_mode(font_color_fg()().a);
}

void CaptionImpl::_handle_theme_change(const std::shared_ptr<Theme>& t)
{
    font_color_bg(t->color_font_bg());
    font_color_fg(t->color_font_fg());
}
