#include <utility>

#include <basic_widgets/core/draw.hpp>
#include <basic_widgets/core/math.hpp>
#include <basic_widgets/w/scrollable_area_impl.hpp>
#include <basic_widgets/w/widget_factory.hpp>

using namespace bwidgets;
using namespace std;

ScrollableAreaImpl::ScrollableAreaImpl(Widget* p)
  : WidgetImpl {p},
    _alignment {Alignment::CENTER_CENTER},
    _scrollbar_h {create_scrollbar(this)},
    _scrollbar_v {create_scrollbar(this)},
    _texture {nullptr},
    _texture_area {}
{
    _scrollbar_h->orientation(ScrollBar::Orientation::HORIZONTAL);
    _scrollbar_v->orientation(ScrollBar::Orientation::VERTICAL);
}

auto ScrollableAreaImpl::alignment() -> Alignment
{
    return _alignment;
}

void ScrollableAreaImpl::alignment(Alignment a)
{
    _alignment = a;
}

void ScrollableAreaImpl::handle_event(const SDL_Event& ev)
{
    WidgetImpl::handle_event(ev);
    _scrollbar_h->handle_event(ev);
    _scrollbar_v->handle_event(ev);
}

auto ScrollableAreaImpl::size() const noexcept -> Size
{
    return {100, 100};
}

auto ScrollableAreaImpl::texture() -> const shared_ptr<Texture>&
{
    return _texture;
}

void ScrollableAreaImpl::texture(shared_ptr<Texture> t)
{
    const auto& attr {t->attributes()};
    _scrollbar_h->step(1.0F / attr.w * 20.F);
    _scrollbar_v->step(1.0F / attr.h * 20.F);
    _texture = move(t);
}

void ScrollableAreaImpl::_handle_geometry_change(const SDL_Rect& vp)
{
    const auto reserved_h = _scrollbar_h->size().w;
    const auto reserved_w = _scrollbar_v->size().w;

    _texture_area = {0, 0, vp.w - reserved_w, vp.h - reserved_h};

    _scrollbar_h->viewport({vp.x, vp.y + _texture_area.h, vp.w, reserved_h});
    _scrollbar_v->viewport(
      {vp.x + _texture_area.w, vp.y, reserved_w, vp.h - reserved_h});
}

void ScrollableAreaImpl::_handle_renderer_change(const shared_ptr<Renderer>& r)
{
    _scrollbar_h->renderer(r);
    _scrollbar_v->renderer(r);
}

void ScrollableAreaImpl::_handle_rendering()
{
    fill_rect_gradient(*renderer(), _texture_area, theme()->color_widget_border(),
                       theme()->color_widget_bg(), theme()->size_widget_border());

    if (_texture) {
        const auto texture_attr = _texture->attributes();
        const auto dst_rect     = [this, &texture_attr]() {
            auto dst = rect_margin(_texture_area, {theme()->size_widget_border(),
                                                   theme()->size_widget_border()});
            if (dst.w > texture_attr.w) {
                switch (alignment()) {
                    case Alignment::BOTTOM_CENTER:
                    case Alignment::CENTER_CENTER:
                    case Alignment::TOP_CENTER:
                        dst.x += center_line(dst.w, texture_attr.w);
                        break;
                    case Alignment::BOTTOM_LEFT:
                    case Alignment::CENTER_LEFT:
                    case Alignment::TOP_LEFT:
                        break;
                    case Alignment::BOTTOM_RIGHT:
                    case Alignment::CENTER_RIGHT:
                    case Alignment::TOP_RIGHT:
                        dst.x += dst.w - texture_attr.w;
                        break;
                }
                dst.w = texture_attr.w;
            }
            if (dst.h > texture_attr.h) {
                switch (alignment()) {
                    case Alignment::BOTTOM_CENTER:
                    case Alignment::BOTTOM_LEFT:
                    case Alignment::BOTTOM_RIGHT:
                        dst.y += dst.h - texture_attr.h;
                        break;
                    case Alignment::CENTER_CENTER:
                    case Alignment::CENTER_LEFT:
                    case Alignment::CENTER_RIGHT:
                        dst.y += center_line(dst.h, texture_attr.h);
                        break;
                    case Alignment::TOP_CENTER:
                    case Alignment::TOP_LEFT:
                    case Alignment::TOP_RIGHT:
                        break;
                }
                dst.h = texture_attr.h;
            }
            return dst;
        }();
        renderer()->copy(
          *_texture,
          {static_cast<int>(_scrollbar_h->value()
                            * clamp(texture_attr.w - dst_rect.w, 0, texture_attr.w)),
           static_cast<int>(_scrollbar_v->value()
                            * clamp(texture_attr.h - dst_rect.h, 0, texture_attr.h)),
           dst_rect.w, dst_rect.h},
          dst_rect);
    }

    _scrollbar_h->render();
    _scrollbar_v->render();
}

void ScrollableAreaImpl::_handle_theme_change(const shared_ptr<Theme>& t)
{
    _scrollbar_h->theme(t);
    _scrollbar_v->theme(t);
}
