
#include <algorithm>
#include <iostream>
#include <memory>
#include <utility>

#include <basic_widgets/core/draw.hpp>
#include <basic_widgets/core/math.hpp>
#include <basic_widgets/w/scrollbar_impl.hpp>
#include <basic_widgets/w/widget_factory.hpp>

using namespace bwidgets;
using namespace std;

auto common_size(const Widget* w1, const Widget* w2) -> Size
{
    const auto size1 = w1->size();
    const auto size2 = w2->size();

    return {size1.w > size2.w ? size1.w : size2.w,
            size1.h > size2.h ? size1.h : size2.h};
}

auto square_size(Size s) -> int
{
    return s.w > s.h ? s.w : s.h;
}

ScrollBarImpl::ScrollBarImpl(Widget* parent)
  : WidgetImpl {parent},
    _button_minus {create_button(this)},
    _button_plus {create_button(this)},
    _orientation {Orientation::VERTICAL},
    _step {.05F}
{
    _button_minus->text("▲");
    _button_plus->text("▼");

    _button_minus->click_handler([this](auto) { value(value() - _step); });
    _button_plus->click_handler([this](auto) { value(value() + _step); });
    click_handler([this](const SDL_MouseButtonEvent ev) {
        SDL_Point p {ev.x - viewport().x - _bar_area.x,
                     ev.y - viewport().y - _bar_area.y};
        if (_orientation == Orientation::VERTICAL) {
            value(p.y / static_cast<float>(_bar_area.h));
        }
        else {
            value(p.x / static_cast<float>(_bar_area.w));
        }
    });
    mouse_motion_handler([this](SDL_MouseMotionEvent ev) {
        if (!pushed()) return;
        SDL_Point p = {ev.x - viewport().x - _bar_area.x,
                       ev.y - viewport().y - _bar_area.y};
        if (_orientation == Orientation::VERTICAL) {
            value(clamp(p.y, 0, _bar_area.h) / static_cast<float>(_bar_area.h));
        }
        else {
            value(clamp(p.x, 0, _bar_area.w) / static_cast<float>(_bar_area.w));
        }
    });
    mouse_wheel_handler([this](SDL_MouseWheelEvent ev) {
        if (orientation() == ScrollBar::Orientation::VERTICAL) {
            if (ev.y < 0) value(value() + _step);
            else if (ev.y > 0) value(value() - _step);
        }
        else {
            if (ev.x > 0) value(value() + _step);
            else if (ev.x < 0) value(value() - _step);
        }
    });
    enable_mouse_handler(_bar_area, viewport());
}

auto ScrollBarImpl::cursor_area() const -> SDL_Rect
{
    if (_orientation == Orientation::VERTICAL) {
        const auto offset = static_cast<int>(value() * _bar_area.h);
        return {_bar_area.x, _bar_area.y + offset - _cursor_size.h / 2, _bar_area.w,
                _cursor_size.h};
    }
    const auto offset = static_cast<int>(value() * _bar_area.w);
    return {_bar_area.x + offset - _cursor_size.h / 2, _bar_area.y, _cursor_size.h,
            _bar_area.h};
}

void ScrollBarImpl::handle_event(const SDL_Event& ev)
{
    WidgetImpl::handle_event(ev);
    _button_minus->handle_event(ev);
    _button_plus->handle_event(ev);
}

void ScrollBarImpl::on_scroll(function<void(float)> cb)
{
    _on_scroll = move(cb);
}

auto ScrollBarImpl::orientation() const -> Orientation
{
    return _orientation;
}

void ScrollBarImpl::orientation(const Orientation o)
{
    switch (o) {
        case Orientation::VERTICAL:
            _button_minus->text("▲");
            _button_plus->text("▼");
            break;
        case Orientation::HORIZONTAL:
            _button_minus->text("◀");
            _button_plus->text("▶");
            break;
    }
    _orientation = o;
}

auto ScrollBarImpl::size() const noexcept -> Size
{
    const auto button_size =
      square_size(common_size(_button_minus.get(), _button_plus.get()));
    if (_orientation == Orientation::VERTICAL) {
        const auto width = [this, button_size]() {
            const auto button_max_width =
              button_size > button_size ? button_size : button_size;
            return button_max_width > _cursor_size.w ? button_max_width : _cursor_size.w;
        }();
        return {width, button_size + button_size + 3 * _cursor_size.h};
    }

    const auto height = button_size > _cursor_size.w ? button_size : _cursor_size.w;
    return {height, button_size + button_size + 3 * _cursor_size.h};
}

auto ScrollBarImpl::step() const -> float
{
    return _step;
}

void ScrollBarImpl::step(float s)
{
    _step = s;
}

auto ScrollBarImpl::value() const -> float
{
    return _value;
}

void ScrollBarImpl::value(const float v)
{
    const auto new_value = clamp(v, 0.F, 1.F);
    if (new_value != _value) {
        _value = new_value;
        if (_on_scroll) _on_scroll(new_value);
    }
}

void ScrollBarImpl::_handle_geometry_change(const SDL_Rect& vp)
{
    const auto button_size =
      square_size(common_size(_button_minus.get(), _button_plus.get()));

    if (_orientation == Orientation::VERTICAL) {
        _bar_area = {0, button_size, vp.w, vp.h - 2 * button_size};
        _button_minus->viewport(rect_offset({0, 0, vp.w, button_size}, vp));
        _button_plus->viewport(
          rect_offset({0, _bar_area.y + _bar_area.h, vp.w, button_size}, vp));
    }
    else {
        _bar_area = {button_size, 0, vp.w - 2 * button_size, vp.h};
        _button_minus->viewport(rect_offset({0, 0, button_size, vp.h}, vp));
        _button_plus->viewport(
          rect_offset({_bar_area.x + _bar_area.w, 0, button_size, vp.h}, vp));
    }
}

void ScrollBarImpl::_handle_renderer_change(const shared_ptr<Renderer>& r)
{
    _button_minus->renderer(r);
    _button_plus->renderer(r);
}

void ScrollBarImpl::_handle_rendering()
{
    fill_rect_gradient(*renderer(), _bar_area, theme()->color_widget_border(),
                       theme()->color_widget_bg_dark());

    auto c = hovered() ? theme()->color_widget_bg_hl() : theme()->color_widget_bg();
    fill_rect_gradient(*renderer(), cursor_area(), theme()->color_widget_border(), c);

    _button_minus->render();
    _button_plus->render();
}

void ScrollBarImpl::_handle_theme_change(const shared_ptr<Theme>& t)
{
    _button_minus->theme(t);
    _button_minus->font(t->font_default_smaller());
    _button_plus->theme(t);
    _button_plus->font(t->font_default_smaller());
}
